#include "avisynth.h"
#include <stdio.h>
#include <Windows.h>
#include <vector>
#include <gdiplus.h>
#include <immintrin.h>
#include <eh.h>

#pragma comment(lib, "gdiplus.lib")

using namespace Gdiplus;

#define BORDER_SIZE 8

void debug(const char* a, ...) {
	char string[1024];
	va_list args;

	va_start(args, a);
	vsprintf_s(string, a, args);
	va_end(args);

	OutputDebugString(string);
}

/***********************************************************************
************************************************************************
* _DCS_GammaDownsize
************************************************************************
***********************************************************************/
class _DCS_GammaDownsize : public GenericVideoFilter {
private:
	int width;
	int height;
	__m128i* columns;
	int* col_widths;
	int* row_heights;

public:
	_DCS_GammaDownsize(PClip _child, int _width, int _height, IScriptEnvironment* env);
	~_DCS_GammaDownsize();
	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};

_DCS_GammaDownsize::_DCS_GammaDownsize(PClip _child, int _width, int _height, IScriptEnvironment* env) : GenericVideoFilter(_child), width(_width), height(_height) {
	if (vi.pixel_type != VideoInfo::CS_BGR32) env->ThrowError("_DCS_GammaDownsize: This filter is for RGB32 input only");
	if (width > vi.width || height > vi.height) env->ThrowError("_DCS_GammaDownsize: This filter is for downsizing only");

	columns = (__m128i*)_aligned_malloc(width * sizeof(__m128i), 16);
	col_widths = new int[width];
	row_heights = new int[height];

// column supersample widths
	int min_col_width = vi.width / width;
	int extras = vi.width - (min_col_width * width);
	int bres = extras - (width >> 1);

	for (int x = 0; x < width; ++x) {
		if (bres >= 0) {
			col_widths[x] = min_col_width + 1;
			bres -= width;
		} else {
			col_widths[x] = min_col_width;
		}

		bres += extras;
	}

// row supersample heights
	int min_row_height = vi.height / height;
	extras = vi.height - (min_row_height * height);
	bres = extras - (height >> 1);

	for (int y = 0; y < height; ++y) {
		if (bres >= 0) {
			row_heights[y] = min_row_height + 1;
			bres -= height;
		} else {
			row_heights[y] = min_row_height;
		}

		bres += extras;
	}

	vi.width = width;
	vi.height = height;
}

_DCS_GammaDownsize::~_DCS_GammaDownsize() {
	_aligned_free(columns);
	delete[] col_widths;
	delete[] row_heights;
}

PVideoFrame __stdcall _DCS_GammaDownsize::GetFrame(int n, IScriptEnvironment* env) {
	PVideoFrame src = child->GetFrame(n, env);
	PVideoFrame dst = env->NewVideoFrame(vi);

	auto src_p = src->GetReadPtr();
	int src_pitch = src->GetPitch();

	auto dst_p = dst->GetWritePtr();
	int dst_pitch = dst->GetPitch();

	__m128i pixels;
	__m128i shuffle = _mm_set_epi32(0x80808003, 0x80808002, 0x80808001, 0x80808000);
	__m128i pack = _mm_set_epi32(0x80808080, 0x80808080, 0x80808080, 0x0c080400);

	int lim_y = 0;
	for (int y = 0; y < vi.height; ++y) {
		ZeroMemory(columns, vi.width << 4);
		for (int src_y = 0; src_y < row_heights[y]; ++src_y) {
			int src_x = 0;
			int lim_x = 0;
			for (int x = 0; x < vi.width; ++x) {
				__m128i total = _mm_set_epi32(0, 0, 0, 0);
				lim_x += col_widths[x];
				for (; src_x < lim_x; ++src_x) {
					if (!(src_x & 3)) {
						pixels = _mm_loadu_si128((__m128i*)&((uint32_t*)src_p)[src_x]);
					} else {
						pixels = _mm_srli_si128(pixels, 4);
					}
					__m128i temp = _mm_shuffle_epi8(pixels, shuffle);
					total = _mm_add_epi32(total, _mm_mullo_epi32(temp, temp));
				}
				columns[x] = _mm_add_epi32(columns[x], total);
			}

			src_p += src_pitch;
		}

		for (int x = 0; x < width; ++x) {
			__m128 div = _mm_set_ps1((float)(col_widths[x] * row_heights[y]));
			((uint32_t*)dst_p)[x] = _mm_cvtsi128_si32(_mm_shuffle_epi8(_mm_cvtps_epi32(_mm_sqrt_ps(_mm_div_ps(_mm_cvtepi32_ps(columns[x]), div))), pack));
		}

		dst_p += dst_pitch;
	}

	return dst;
}

/***********************************************************************
************************************************************************
* DebugColourSpaces
************************************************************************
***********************************************************************/
class DebugColourSpaces : public GenericVideoFilter {
private:
	int	width;
	int max_height;
	uint8_t* bitmap;
	ULONG_PTR m_gdiplusToken = NULL;
	Pen* outline;
	Brush* white;
	Brush* cyan;
	Brush* yellow;
	Brush* orange;
	Brush* red;
	Brush* half_yellow;
	Brush* half_orange;
	Brush* half_red;
	Brush* shadow;
	float font_size;
	FontFamily* font_family;
	StringFormat* string_format;
	GraphicsPath* space_names_path;
	GraphicsPath* conv_names_path;
	GraphicsPath* boxes;
	GraphicsPath* cross;
	wchar_t temp_text[64];

	int depths[6] = { 8, 10, 12, 14, 16, 32 };
	std::vector<PClip> depth_clips;

	struct ColourSpace {
		const wchar_t* name;
		const char* converter;
		int depth;
		bool alpha;
		bool implicit_alpha_conversion;
		bool implicit_depth_conversion;
		int pixel_type;
		VideoInfo output_vi;
		int output_space;

		PClip clip;
		enum error_type { NONE = 0, AVS_ERROR, EXCEPTION } error = NONE;
		GraphicsPath* error_path = NULL;
	};

	std::vector<ColourSpace> spaces{
		{ L"Y8",              "ConvertToY",          0, false, false, false, VideoInfo::CS_Y8 },
		{ L"YUV420P8 (YV12)", "ConvertToYUV420",     0, false, false, false, VideoInfo::CS_YV12 },
		{ L"YUV422P8 (YV16)", "ConvertToYUV422",     0, false, false, false, VideoInfo::CS_YV16 },
		{ L"YUV444P8 (YV24)", "ConvertToYUV444",     0, false, false, false, VideoInfo::CS_YV24 },
		{ L"RGBP8",           "ConvertToPlanarRGB",  0, false, false, false, VideoInfo::CS_RGBP8 },
		{ L"YUVA420P8",       "ConvertToYUV420",     0, true,  false, false, VideoInfo::CS_YUVA420 },
		{ L"YUVA422P8",       "ConvertToYUV422",     0, true,  false, false, VideoInfo::CS_YUVA422 },
		{ L"YUVA444P8",       "ConvertToYUV444",     0, true,  false, false, VideoInfo::CS_YUVA444 },
		{ L"RGBAP8",          "ConvertToPlanarRGBA", 0, true,  false, false, VideoInfo::CS_RGBAP8 },

		{ L"Y10",             "ConvertToY",          1, false, false, false, VideoInfo::CS_Y10 },
		{ L"YUV420P10",       "ConvertToYUV420",     1, false, false, false, VideoInfo::CS_YUV420P10 },
		{ L"YUV422P10",       "ConvertToYUV422",     1, false, false, false, VideoInfo::CS_YUV422P10 },
		{ L"YUV444P10",       "ConvertToYUV444",     1, false, false, false, VideoInfo::CS_YUV444P10 },
		{ L"RGBP10",          "ConvertToPlanarRGB",  1, false, false, false, VideoInfo::CS_RGBP10 },
		{ L"YUVA420P10",      "ConvertToYUV420",     1, true,  false, false, VideoInfo::CS_YUVA420P10 },
		{ L"YUVA422P10",      "ConvertToYUV422",     1, true,  false, false, VideoInfo::CS_YUVA422P10 },
		{ L"YUVA444P10",      "ConvertToYUV444",     1, true,  false, false, VideoInfo::CS_YUVA444P10 },
		{ L"RGBAP10",         "ConvertToPlanarRGBA", 1, true,  false, false, VideoInfo::CS_RGBAP10 },

		{ L"Y12",             "ConvertToY",          2, false, false, false, VideoInfo::CS_Y12 },
		{ L"YUV420P12",       "ConvertToYUV420",     2, false, false, false, VideoInfo::CS_YUV420P12 },
		{ L"YUV422P12",       "ConvertToYUV422",     2, false, false, false, VideoInfo::CS_YUV422P12 },
		{ L"YUV444P12",       "ConvertToYUV444",     2, false, false, false, VideoInfo::CS_YUV444P12 },
		{ L"RGBP12",          "ConvertToPlanarRGB",  2, false, false, false, VideoInfo::CS_RGBP12 },
		{ L"YUVA420P12",      "ConvertToYUV420",     2, true,  false, false, VideoInfo::CS_YUVA420P12 },
		{ L"YUVA422P12",      "ConvertToYUV422",     2, true,  false, false, VideoInfo::CS_YUVA422P12 },
		{ L"YUVA444P12",      "ConvertToYUV444",     2, true,  false, false, VideoInfo::CS_YUVA444P12 },
		{ L"RGBAP12",         "ConvertToPlanarRGBA", 2, true,  false, false, VideoInfo::CS_RGBAP12 },

		{ L"Y14",             "ConvertToY",          3, false, false, false, VideoInfo::CS_Y14 },
		{ L"YUV420P14",       "ConvertToYUV420",     3, false, false, false, VideoInfo::CS_YUV420P14 },
		{ L"YUV422P14",       "ConvertToYUV422",     3, false, false, false, VideoInfo::CS_YUV422P14 },
		{ L"YUV444P14",       "ConvertToYUV444",     3, false, false, false, VideoInfo::CS_YUV444P14 },
		{ L"RGBP14",          "ConvertToPlanarRGB",  3, false, false, false, VideoInfo::CS_RGBP14 },
		{ L"YUVA420P14",      "ConvertToYUV420",     3, true,  false, false, VideoInfo::CS_YUVA420P14 },
		{ L"YUVA422P14",      "ConvertToYUV422",     3, true,  false, false, VideoInfo::CS_YUVA422P14 },
		{ L"YUVA444P14",      "ConvertToYUV444",     3, true,  false, false, VideoInfo::CS_YUVA444P14 },
		{ L"RGBAP14",         "ConvertToPlanarRGBA", 3, true,  false, false, VideoInfo::CS_RGBAP14 },

		{ L"Y16",             "ConvertToY",          4, false, false, false, VideoInfo::CS_Y16 },
		{ L"YUV420P16",       "ConvertToYUV420",     4, false, false, false, VideoInfo::CS_YUV420P16 },
		{ L"YUV422P16",       "ConvertToYUV422",     4, false, false, false, VideoInfo::CS_YUV422P16 },
		{ L"YUV444P16",       "ConvertToYUV444",     4, false, false, false, VideoInfo::CS_YUV444P16 },
		{ L"RGBP16",          "ConvertToPlanarRGB",  4, false, false, false, VideoInfo::CS_RGBP16 },
		{ L"YUVA420P16",      "ConvertToYUV420",     4, true,  false, false, VideoInfo::CS_YUVA420P16 },
		{ L"YUVA422P16",      "ConvertToYUV422",     4, true,  false, false, VideoInfo::CS_YUVA422P16 },
		{ L"YUVA444P16",      "ConvertToYUV444",     4, true,  false, false, VideoInfo::CS_YUVA444P16 },
		{ L"RGBAP16",         "ConvertToPlanarRGBA", 4, true,  false, false, VideoInfo::CS_RGBAP16 },

		{ L"Y32",             "ConvertToY",          5, false, false, false, VideoInfo::CS_Y32 },
		{ L"YUV420PS",        "ConvertToYUV420",     5, false, false, false, VideoInfo::CS_YUV420PS },
		{ L"YUV422PS",        "ConvertToYUV422",     5, false, false, false, VideoInfo::CS_YUV422PS },
		{ L"YUV444PS",        "ConvertToYUV444",     5, false, false, false, VideoInfo::CS_YUV444PS },
		{ L"RGBPS",           "ConvertToPlanarRGB",  5, false, false, false, VideoInfo::CS_RGBPS },
		{ L"YUVA420PS",       "ConvertToYUV420",     5, true,  false, false, VideoInfo::CS_YUVA420PS },
		{ L"YUVA422PS",       "ConvertToYUV422",     5, true,  false, false, VideoInfo::CS_YUVA422PS },
		{ L"YUVA444PS",       "ConvertToYUV444",     5, true,  false, false, VideoInfo::CS_YUVA444PS },
		{ L"RGBAPS",          "ConvertToPlanarRGBA", 5, true,  false, false, VideoInfo::CS_RGBAPS },

		{ L"YV411",           "ConvertToYV411",      0, false, true,  false, VideoInfo::CS_YV411 },
		{ L"YUY2",            "ConvertToYUY2",       0, false, true,  false, VideoInfo::CS_YUY2 },

		{ L"RGB24",           "ConvertToRGB24",      0, false, true,  true,  VideoInfo::CS_BGR24 },
		{ L"RGB32",           "ConvertToRGB32",      0, true,  true,  true,  VideoInfo::CS_BGR32 },
		{ L"RGB48",           "ConvertToRGB48",      4, false, true,  true,  VideoInfo::CS_BGR48 },
		{ L"RGB64",           "COnvertToRGB64",      4, true,  true,  true,  VideoInfo::CS_BGR64 }
	};

	struct Layout {
		int x;
		int y;
		std::vector<int> spaces;
	};

	int layout_id;
	std::vector<Layout> layouts{
		{ 1,1, {
		}},
		{ 10,6, {
			54,  0,  1, 2, 3, 4, 5, 6, 7, 8,
			55,  9, 10,11,12,13,14,15,16,17,
			56, 18, 19,20,21,22,23,24,25,26,
			57, 27, 28,29,30,31,32,33,34,35,
			58, 36, 37,38,39,40,41,42,43,44,
			59, 45, 46,47,48,49,50,51,52,53
		}},
		{ 4,2, {
			 0,  1,  2,  3,
			54, 55, 56, 57
		}},
		{ 5,6, {
			54,  0, 1, 2, 3,
			55,  9,10,11,12,
			-1, 18,19,20,21,
			-1, 27,28,29,30,
			-1, 36,37,38,39,
			-1, 45,46,47,48
		}},
		{ 5,2, {
			  4,13,22,31,40,
			 56,57,58,59,49
		}}
	};

public:
	DebugColourSpaces(PClip _child, int display_width, int _layout_id, bool interlaced, const char* filter, const AVSValue args, int args_start, IScriptEnvironment* env);
	~DebugColourSpaces();
	GraphicsPath* ErrorPath(const char* text);
	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};

/***********************************************************************
* ErrorPath
***********************************************************************/
GraphicsPath* DebugColourSpaces::ErrorPath(const char* text) {
	float font_size = this->font_size;
	GraphicsPath* path = new GraphicsPath();
	int target_width = (int)(width * 0.8);
	
	string_format->SetAlignment(StringAlignment::StringAlignmentCenter);
	string_format->SetLineAlignment(StringAlignment::StringAlignmentCenter);

	static wchar_t temp[256];
	static RectF rect;
	swprintf_s(temp, 1024, L"%S", text);

// replace all spaces with newlines
	int p = 0;
	while (temp[p]) {
		if (temp[p] == L' ') temp[p] = L'\n';
		++p;
	}

	path->AddString(temp, -1, font_family, FontStyleRegular, font_size, PointF(0, 0), string_format);
	path->GetBounds(&rect);

	if (rect.Width > target_width) font_size *= (float)target_width / rect.Width;

	p = 0;
	bool okay = false;
	while (true) {
		while (temp[p] && temp[p] != L'\n') ++p;
		if (temp[p]) temp[p] = L' ';

		if (temp[p] || !okay) {
			path->Reset();
			path->AddString(temp, -1, font_family, FontStyleRegular, font_size, PointF(0, 0), string_format);
			path->GetBounds(&rect);
		}

		if (!temp[p]) break;

		if (rect.Width > target_width) {
			temp[p] = L'\n';
			okay = false;
		} else {
			okay = true;
		}

		++p;
	}

	return path;
}

/***********************************************************************
* Exception translator
***********************************************************************/
class StackOverflowException{};

void _DCS_TransFunc(unsigned int u, EXCEPTION_POINTERS* pExc) {
	if (u == EXCEPTION_STACK_OVERFLOW) throw StackOverflowException();
	static char message[256];

	const EXCEPTION_RECORD& exr = *pExc->ExceptionRecord;
	switch (u) {
		case EXCEPTION_ACCESS_VIOLATION: sprintf_s(message, "Access violation %s location 0x%p", exr.ExceptionInformation[0] ? "writing" : "reading", (void*)exr.ExceptionInformation[1]); break;
		case EXCEPTION_INT_DIVIDE_BY_ZERO: sprintf_s(message, "Integer division by zero"); break;
		case EXCEPTION_FLT_DIVIDE_BY_ZERO: sprintf_s(message, "Floating point division by zero"); break;
		case EXCEPTION_DATATYPE_MISALIGNMENT: sprintf_s(message, "Data type misalignment"); break;
		case EXCEPTION_ILLEGAL_INSTRUCTION: sprintf_s(message, "Illegal instruction"); break;
		case EXCEPTION_FLT_INVALID_OPERATION: sprintf_s(message, "Floating point invalid operation"); break;
		case EXCEPTION_INT_OVERFLOW: sprintf_s(message, "Integer overflow"); break;
	}

	throw std::exception(message);
}

/***********************************************************************
************************************************************************
* Constructor
************************************************************************
***********************************************************************/
DebugColourSpaces::DebugColourSpaces(PClip _child, int display_width, int _layout_id, bool interlaced, const char* filter, const AVSValue args, int args_start, IScriptEnvironment* env) : GenericVideoFilter(_child), layout_id(_layout_id) {
	if (!env->FunctionExists(filter)) env->ThrowError("DebugColourSpaces: Function '%s' does not exist", filter);

	if (layout_id < 0 || layout_id >= (int)layouts.size()) env->ThrowError("DebugColourSpaces: Bad layout ID");

	if (layout_id == 0) {
		int src = -1;
		for (int s = 0; s < (int)spaces.size(); ++s) {
			if (spaces[s].pixel_type == vi.pixel_type) {
				src = s;
			}
		}
		if (src == -1) env->ThrowError("DebugColourSpaces: Could not find source pixel_type (this should not happen!)");
		layouts[0].spaces.push_back(src);
	}

// Can't call ConvertBits on interleaved video, so first convert to planar if necessary
	PClip planar_child = NULL;

	if (!vi.IsPlanar()) {
		if (vi.IsRGB24() || vi.IsRGB48()) {
			planar_child = env->Invoke("ConvertToPlanarRGB", AVSValue(AVSValue(child), 1)).AsClip();
		} else if (vi.IsRGB32() || vi.IsRGB64()) {
			planar_child = env->Invoke("ConvertToPlanarRGBA", AVSValue(AVSValue(child), 1)).AsClip();
		} else {
			planar_child = env->Invoke("ConvertToYV16", AVSValue(AVSValue(child), 1)).AsClip();
		}
	} else if (vi.IsYV411()) {
		planar_child = env->Invoke("ConvertToYUV422", AVSValue(AVSValue(child), 1)).AsClip();
	}

// Convert to every bit depth
	for (auto& depth : depths) {
		if (vi.BitsPerComponent() == depth) {
			depth_clips.push_back(!vi.IsPlanar() ? planar_child : child); // changed from planar_child ? ... because YV411 is converted above, but can still be put in depth_clips
		} else {
			try {
				AVSValue args[2] = { planar_child ? planar_child : child, depth };
				PClip temp = env->Invoke("ConvertBits", AVSValue(args, 2)).AsClip();
				depth_clips.push_back(temp);
			} catch (AvisynthError& e) {
				env->ThrowError("DebugColourSpaces (depths): %s", e.msg);
			} catch (...) {
				env->ThrowError("DebugColourSpaces: Something went wrong which should not have gone wrong! (depths)");
			}
		}
	}

// Initiate filter args
	int n_args = args.ArraySize() - args_start;
	auto filter_args = std::vector<AVSValue>(n_args + 1);
	for (int i = 0; i < n_args; ++i) {
		filter_args[i + 1] = args[i + args_start];
	}

	auto& layout = layouts[layout_id];

// Calculate display width
	if (display_width == -1) {
		width = vi.width;
	} else {
		width = (std::min)(vi.width, (display_width - (layout.x - 1) * BORDER_SIZE) / layout.x);
		if (width < 16) env->ThrowError("DebugColourSpaces: Output width is too small (or negative)");
	}
	display_width = width * layout.x + (layout.x - 1) * BORDER_SIZE;
	max_height = 0;
	int max_frames = 1;

// gdiplus setup
	GdiplusStartupInput gdiplusStartupInput;
	GdiplusStartup(&m_gdiplusToken, &gdiplusStartupInput, NULL);

// pens and brushes
	outline = new Pen(Color(0, 0, 0), 3);
	outline->SetLineJoin(LineJoinRound);
	white = new SolidBrush(Color(255, 255, 255));
	cyan = new SolidBrush(Color(128, 255, 255));

	yellow = new SolidBrush(Color(255, 255, 0));
	orange = new SolidBrush(Color(255, 128, 0));
	red = new SolidBrush(Color(255, 0, 0));
	half_yellow = new SolidBrush(Color(16, 255, 255, 0));
	half_orange = new SolidBrush(Color(16, 255, 128, 0));
	half_red = new SolidBrush(Color(16, 255, 0, 0));

	shadow = new SolidBrush(Color(192, 0, 0, 0));

// string format
	string_format = new StringFormat();

// paths
	cross = new GraphicsPath();
	boxes = new GraphicsPath();
	space_names_path = new GraphicsPath();
	conv_names_path = new GraphicsPath();

// main font setup
	font_family = new FontFamily(L"Calibri");
	if (!font_family->IsAvailable()) {
		delete font_family;
		font_family = (FontFamily*)FontFamily::GenericSansSerif();
	}

	{
		GraphicsPath temp;
		RectF temp_rect;
		temp.AddString(L"N", -1, font_family, FontStyleRegular, 100, PointF(0, 0), string_format);
		temp.GetBounds(&temp_rect);
		font_size = (width / 14.0f) * 64 / temp_rect.Height; // (std::min)(width, max_height) / 10.0f;
	}

	float box_margin = font_size / 4;
	string_format->SetAlignment(StringAlignment::StringAlignmentCenter);
	string_format->SetLineAlignment(StringAlignment::StringAlignmentCenter);

// Generate all clips
	for (auto& space_id : layout.spaces) {
		if (space_id < 0) continue;
		auto& space = spaces[space_id];

		PClip temp = NULL;
		try {
			if (!_stricmp(space.converter, "ConvertToY")) {
				AVSValue args[1] = { space.implicit_depth_conversion && (vi.pixel_type != VideoInfo::CS_YUY2 || space.depth == 0) ? child : depth_clips[space.depth] };
				temp = env->Invoke(space.converter, AVSValue(args, 1)).AsClip();
			} else {
				AVSValue args[2] = { space.implicit_depth_conversion && (vi.pixel_type != VideoInfo::CS_YUY2 || space.depth == 0) ? child : depth_clips[space.depth], interlaced };
				const char* arg_names[2] = { NULL, "interlaced" };
				temp = env->Invoke(space.converter, AVSValue(args, 2), arg_names).AsClip();
			}
		} catch (AvisynthError& e) {
			env->ThrowError("DebugColourSpaces (conversion): %s", e.msg);
		} catch (...) {
			env->ThrowError("DebugColourSpaces: Something went wrong which should not have gone wrong! (conversion)");
		}

		if (!space.implicit_alpha_conversion && space.alpha != (vi.NumComponents() == 4)) {
			try {
				AVSValue args[1] = { temp };
				temp = env->Invoke(space.alpha ? "AddAlphaPlane" : "RemoveAlphaPlane", AVSValue(args, 1)).AsClip();
			} catch (AvisynthError& e) {
				env->ThrowError("DebugColourSpaces: (add/remove alpha): %s", e.msg);
			} catch (...) {
				env->ThrowError("DebugColourSpaces: Something went wrong which should not have gone wrong! (add/remove alpha)");
			}
		}

		auto restore = _set_se_translator(_DCS_TransFunc);

		bool stack_reset = false;
		bool error = false;

		debug("Invoking %s on %ls...", filter, space.name);

		try {
			filter_args[0] = temp;
			temp = env->Invoke(filter, AVSValue(&filter_args[0], n_args + 1)).AsClip();
			debug("Success!");
			space.output_vi = temp->GetVideoInfo();
		} catch (IScriptEnvironment::NotFound) {
			env->ThrowError("DebugColourSpaces: Invalid arguments to function '%s'", filter);
		} catch (AvisynthError e) {
			space.error_path = ErrorPath(e.msg);
			space.error = ColourSpace::error_type::AVS_ERROR;
			error = true;
		} catch (std::exception& e) {
			space.error_path = ErrorPath(e.what());
			space.error = ColourSpace::error_type::EXCEPTION;
			error = true;
		} catch (StackOverflowException&) {
			stack_reset = true;
		} catch (...) {
			space.error_path = ErrorPath("Unknown exception");
			space.error = ColourSpace::error_type::EXCEPTION;
			error = true;
		}

		if (stack_reset) {
			debug("Stack overflow; attempting reset");
			if (!_resetstkoflw()) {
				debug("Reset failed");
				env->ThrowError("DebugColourSpaces: Failed to recover from a stack overflow");
			}
			space.error_path = ErrorPath("Stack overflow");
			space.error = ColourSpace::error_type::EXCEPTION;
			error = true;
		}

		_set_se_translator(restore);

		if (error) {
			space.clip = NULL;
			continue;
		}

// convert result to RGB32 for display
		AVSValue args[2] = { temp, interlaced };
		const char* arg_names[2] = { NULL, "interlaced" };
		temp = env->Invoke("ConvertToRGB32", AVSValue(args, 1), arg_names).AsClip();

		auto& vi = temp->GetVideoInfo();
		if (vi.width > width) {
			int target_height = (int)(((float)width / vi.width) * vi.height);
			target_height = (std::max)(1, target_height);
			max_height = (std::max)(max_height, target_height);

			AVSValue args[3] = { temp, width, target_height };
			temp = env->Invoke("_DCS_GammaDownsize", AVSValue(args, 3)).AsClip();
		} else {
			max_height = (std::max)(max_height, vi.height);
		}

		max_frames = (std::max)(max_frames, vi.num_frames);
		space.clip = temp;
	}

	if (!max_height) {
		max_height = (std::max)((int)(((float)width / vi.width) * vi.height), 1);
	}

// create cross path
	float scale = (float)(std::min)(width, max_height);
	float a = scale * 0.1f;
	float b = scale * 0.2f;
	float c = a + b;

	PointF points[] = {
		PointF(0, a),
		PointF(b, c),
		PointF(c, b),
		PointF(a, 0),
		PointF(c, -b),
		PointF(b, -c),
		PointF(0, -a),
		PointF(-b, -c),
		PointF(-c, -b),
		PointF(-a, 0),
		PointF(-c, b),
		PointF(-b, c)
	};
	cross->AddPolygon(points, 12);

// Add borders to reach max_height and find output colour space
	int x, y;
	x = y = 0;
	string_format->SetLineAlignment(StringAlignment::StringAlignmentFar);

	for (auto& space_id : layout.spaces) {
		if (space_id >= 0) {
			auto& space = spaces[space_id];

			if (!space.clip) {
				space.output_space = -2;
			} else {
				auto& vi = space.clip->GetVideoInfo();
				if (vi.height < max_height || vi.width < width) {
					AVSValue args[5] = { space.clip, (width - vi.width) >> 1, (max_height - vi.height) >> 1, (width + 1 - vi.width) >> 1, (max_height + 1 - vi.height) >> 1 };
					space.clip = env->Invoke("AddBorders", AVSValue(args, 5)).AsClip();
				}

				if (space.output_vi.pixel_type != space.pixel_type) {
					space.output_space = -1;

					for (int s = 0; s < (int)spaces.size(); ++s) {
						if (space.output_vi.pixel_type == spaces[s].pixel_type) {
							space.output_space = s;
							break;
						}
					}
				} else {
					space.output_space = -2;
				}
			}

// set graphical position
			int space_x = x * (width + BORDER_SIZE);
			int space_y = y * (max_height + BORDER_SIZE);

// add input/output colour space name to path(s)
			GraphicsPath temp;
			string_format->SetAlignment(space.output_space == -2 ? StringAlignment::StringAlignmentCenter : StringAlignment::StringAlignmentNear);
			temp.AddString(space.name, -1, font_family, FontStyleRegular, font_size, Gdiplus::PointF(space_x + (space.output_space == -2 ? width * 0.5f : 0.0f), (float)(space_y + max_height)), string_format);
			space_names_path->AddPath(&temp, false);

			RectF temp_rect;
			temp.GetBounds(&temp_rect);
			temp_rect.X -= box_margin;
			temp_rect.Y -= box_margin;
			temp_rect.Width += box_margin * 2;
			temp_rect.Height += box_margin * 2;
			boxes->AddRectangle(temp_rect);

			if (space.output_space != -2) {
				GraphicsPath temp;
				string_format->SetAlignment(StringAlignment::StringAlignmentFar);
				swprintf_s(temp_text, 64, L"-> %s", space.output_space >= 0 ? spaces[space.output_space].name : L"???");
				temp.AddString(temp_text, -1, font_family, FontStyleRegular, font_size, Gdiplus::PointF((float)(space_x + width), (float)(space_y + max_height)), string_format);
				conv_names_path->AddPath(&temp, false);

				RectF temp_rect;
				temp.GetBounds(&temp_rect);
				temp_rect.X -= box_margin;
				temp_rect.Y -= box_margin;
				temp_rect.Width += box_margin * 2;
				temp_rect.Height += box_margin * 2;
				boxes->AddRectangle(temp_rect);
			}
		}

		++x;
		if (x == layout.x) {
			++y;
			x = 0;
		}
	}

// set video parameters
	vi.width = display_width;
	vi.height = max_height * layout.y + (layout.y - 1) * BORDER_SIZE;
	vi.num_frames = max_frames;
	vi.pixel_type = VideoInfo::CS_BGR32;
	vi.audio_samples_per_second = 0;
	vi.nchannels = 0;
}

/***********************************************************************
************************************************************************
* Destructor
************************************************************************
***********************************************************************/
DebugColourSpaces::~DebugColourSpaces() {
	delete outline;
	delete white;
	delete cyan;

	delete yellow;
	delete orange;
	delete red;
	delete half_yellow;
	delete half_orange;
	delete half_red;

	delete shadow;
	delete font_family;

	delete string_format;
	delete cross, boxes, space_names_path, conv_names_path;

	for (auto& space : spaces) {
		delete space.error_path;
	}

	if (m_gdiplusToken) Gdiplus::GdiplusShutdown(m_gdiplusToken);
}

/***********************************************************************
************************************************************************
* GetFrame
************************************************************************
***********************************************************************/
PVideoFrame __stdcall DebugColourSpaces::GetFrame(int n, IScriptEnvironment* env) {
	debug("DebugColourSpaces: GetFrame");
	PVideoFrame dst = env->NewVideoFrame(vi);
	auto dst_p = dst->GetWritePtr();
	int dst_pitch = dst->GetPitch();
	bool exception;

	dst_p += dst_pitch * (vi.height - 1);
	dst_pitch = -dst_pitch;

// Create graphics context
	Bitmap* bitmap = new Bitmap(vi.width, vi.height, dst_pitch, PixelFormat32bppRGB, dst_p);
	Graphics* g = Graphics::FromImage(bitmap);

	g->SetSmoothingMode(SmoothingModeHighQuality);
	g->SetCompositingQuality(CompositingQuality::CompositingQualityGammaCorrected);
	g->SetTextRenderingHint(TextRenderingHint::TextRenderingHintAntiAliasGridFit);

	auto& layout = layouts[layout_id];

// Paint frames
	int x = 0;
	int y = 0;

	for (auto& space_id : layout.spaces) {
		int space_x = x * (width + BORDER_SIZE);
		int space_y = y * (max_height + BORDER_SIZE);

		auto dst_pp = dst_p + ((space_x << 2) + space_y * dst_pitch);

		if (space_id < 0) {
			for (int y = 0; y < max_height; y++) {
				ZeroMemory(dst_pp, width << 2);
				dst_pp += dst_pitch;
			}
		} else {
			auto& space = spaces[space_id];
			exception = false;

			debug("%ls", space.name);

			if (space.clip) {
				debug("Getting frame", space.name);

				PVideoFrame src;
				bool stack_reset = false;

				auto restore = _set_se_translator(_DCS_TransFunc);

				try {
					src = space.clip->GetFrame(n, env);
					debug("Success");
				} catch (AvisynthError e) {
					space.error_path = ErrorPath(e.msg);
					exception = true;
				} catch (std::exception& e) {
					space.error_path = ErrorPath(e.what());
					exception = true;
				} catch (StackOverflowException&) {
					stack_reset = true;
				} catch (...) {
					space.error_path = ErrorPath("Unknown exception");
					exception = true;
				}

				if (stack_reset) {
					debug("Stack overflow; attempting reset");
					if (!_resetstkoflw()) {
						debug("Reset failed");
						env->ThrowError("DebugColourSpaces: Failed to recover from a stack overflow");
					}
					space.error_path = ErrorPath("Stack overflow");
					exception = true;
				}

				_set_se_translator(restore);

				if (!exception) {
					auto src_p = src->GetReadPtr();
					int src_pitch = src->GetPitch();
					src_p += src_pitch * (src->GetHeight() - 1);
					src_pitch = -src_pitch;

					env->BitBlt(dst_pp, dst_pitch, src_p, src_pitch, src->GetRowSize(), src->GetHeight());
				}
			}

			if (!space.clip || exception) {
				debug("no clip or exception");
				for (int y = 0; y < max_height; y++) {
					for (int x = 0; x < width; x++) {
						((uint32_t*)dst_pp)[x] = 0xff202020;
					}
					dst_pp += dst_pitch;
				}

				g->SetClip(Rect(space_x, space_y, width, max_height));
				g->TranslateTransform(space_x + width * 0.5f, space_y + max_height * 0.5f);
				Brush* col;
				Brush* half_col;
				switch (space.error) {
					case ColourSpace::error_type::AVS_ERROR: col = yellow; half_col = half_yellow; break;
					case ColourSpace::error_type::EXCEPTION: col = red; half_col = half_red; break;
					default: col = orange; half_col = half_orange; break;
				}
				g->FillPath(half_col, cross);
				if (space.error_path) {
					g->DrawPath(outline, space.error_path);
					g->FillPath(col, space.error_path);
				}
				g->ResetTransform();
				g->ResetClip();
			}
		}

		if (x < layout.x - 1) {
			dst_pp = dst_p + (y * (max_height + BORDER_SIZE)) * dst_pitch + ((x * (width + BORDER_SIZE) + width) << 2);
			for (int by = 0; by < max_height; ++by) {
				ZeroMemory(dst_pp, BORDER_SIZE << 2);
				dst_pp += dst_pitch;
			}
		}

		++x;
		if (x == layout.x) {
			if (y < layout.y - 1) {
				dst_pp = dst_p + (y * (max_height + BORDER_SIZE) + max_height) * dst_pitch;
				for (int by = 0; by < BORDER_SIZE; ++by) {
					ZeroMemory(dst_pp, -dst_pitch);
					dst_pp += dst_pitch;
				}
			}

			++y;
			x = 0;
			// fill column
		}
	}

	g->FillPath(shadow, boxes);

	g->DrawPath(outline, space_names_path);
	g->DrawPath(outline, conv_names_path);

	g->FillPath(white, space_names_path);
	g->FillPath(cyan, conv_names_path);

	delete bitmap;
	delete g;

	return dst;
}

AVSValue __cdecl Create_DCS_GammaDownsize(AVSValue args, void* user_data, IScriptEnvironment* env) {
	return new _DCS_GammaDownsize(
		args[0].AsClip(),
		args[1].AsInt(),
		args[2].AsInt(),
		env
	);
}

AVSValue __cdecl Create_DebugColourSpaces(AVSValue args, void* user_data, IScriptEnvironment* env) {
	int i = 0;
	PClip input = NULL;
	bool got_width = false;
	bool got_layout = false;
	bool got_interlaced = false;
	int width = GetSystemMetrics(SM_CXSCREEN);
	int layout_id = 1;
	bool interlaced = false;

	while (i < args[0].ArraySize() && !args[0][i].IsString()) {
		if (args[0][i].IsClip()) {
			if (input) env->ThrowError("Invalid arguments to function 'DebugColourSpaces': multiple clips");
			input = args[0][i++].AsClip();
		} else if (args[0][i].IsInt()) {
			if (got_width) env->ThrowError("Invalid arguments to function 'DebugColourSpaces' (bad parameters)");
			width = args[0][i++].AsInt();
			got_width = true;
		} else if (args[0][i].IsFloat()) {
			float temp = args[0][i++].AsFloatf();
			if (got_layout || (temp - floorf(temp) != 0.5f)) env->ThrowError("Invalid arguments to function 'DebugColourSpaces' (bad parameters)");
			layout_id = (int)(floorf(temp - 0.5f));
			got_layout = true;
		} else if (args[0][i].IsBool()) {
			if (got_interlaced) env->ThrowError("Invalid arguments to function 'DebugColourSpaces' (bad parameters)");
			interlaced = args[0][i++].AsBool();
			got_interlaced = true;
		} else {
			if (got_width) env->ThrowError("Invalid arguments to function 'DebugColourSpaces' (bad parameters)");
		}
	}

	if (!input) input = env->GetVarDef("last", 0).AsClip();
	if (!input) env->ThrowError("Invalid arguments to function 'DebugColourSpaces' (no input clip specified)");

	if (i >= args[0].ArraySize() || !args[0][i].IsString()) env->ThrowError("Invalid arguments to function `DebugColourSpaces` (no filter name specified)");

	return new DebugColourSpaces(
		input,
		width,
		layout_id,
		interlaced,
		args[0][i].AsString(),
		args[0],
		i + 1,
		env
	);
}

const AVS_Linkage* AVS_linkage = 0;

AVSValue __cdecl Create_half_const(AVSValue args, void* user_data, IScriptEnvironment* env) {
	return AVSValue((float)((int64_t)user_data + 0.5f));
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit3(IScriptEnvironment* env, const AVS_Linkage* const vectors) {
	AVS_linkage = vectors;

	env->AddFunction("_DCS_GammaDownsize", "cii", Create_DCS_GammaDownsize, 0);
	env->AddFunction("DebugColorSpaces", ".+", Create_DebugColourSpaces, 0);
	env->AddFunction("DebugColourSpaces", ".+", Create_DebugColourSpaces, 0);
	env->AddFunction("DCS_SOURCE", "", Create_half_const, (void*)0);
	env->AddFunction("DCS_ALL", "", Create_half_const, (void*)1);
	env->AddFunction("DCS_CLASSIC", "", Create_half_const, (void*)2);
	env->AddFunction("DCS_YUV", "", Create_half_const, (void*)3);
	env->AddFunction("DCS_RGB", "", Create_half_const, (void*)4);

	return "DebugColourSpaces";
}
